/******************************************************************************
 * $Header: /boot/home/agmsmith/Programming/WordWrapper/RCS/AGMSWordWrapper.cpp,v 1.8 2002/10/28 23:59:25 agmsmith Exp $
 *
 * This is a BeIDE text editor add-on which wraps the selected text to fit in a
 * width of 79 characters.  It requires the BeIDE which comes with BeOS 5.0.
 * It also understands UTF-8 character encoding, block comments, C style
 * comments, recognizes certain abbreviations, tab stops, and adds extra space
 * between sentences.
 *
 * To use it, select the paragraph of text you want wrapped.  All the lines
 * that are partially or fully inside the selected area will be wrapped (so you
 * can be sloppy about making your selection).
 *
 * The indent for the wrapping is taken from the first line; all wrapped lines
 * will have a copy of that indent text in front of them.  The indent text
 * consists of all initial space or punctuation characters of the first line.
 * An exception is made for the start of C style comments (slash-star), which
 * won't be included (otherwise all lines will start with slash-star).  So, you
 * can have " * " as the indent for block style comments (like this comment).
 * Tabs in the indent count as 1 space, because I don't know how to access the
 * editor's tab settings, and I use spaces anyways so that my code looks
 * correct no matter what the tab stops are set to.  If you want to change
 * that, I recommend that you convert the tabs into multiple spaces before the
 * main part of the processing, and convert back to tabs afterwards.  The
 * remainder of the text will have the indent removed (as much as matches the
 * first line's indent), and be word wrapped (spaces separate the words) to fit
 * in the remainder of the space up to 79 characters in line width, with the
 * indent text re-inserted at the beginning of each new line.
 *
 * If the initial white space is too big (more than 79 characters) it will do
 * nothing.  The text is interpreted as UTF-8 encoded characters, so multibyte
 * symbols count as one character.  Also, multiple white space characters
 * (spaces, tabs, etc) in the wrapped text get removed and replaced by a single
 * space, except at the ends of sentences (after periods, question marks and
 * exclamation marks where they get two spaces between them and the next word,
 * except for single letter abbreviations (but not single digit and other
 * single non-letter ones) and the end of a C style comment).  There are no
 * trailing spaces at the ends of lines.  Words longer than the maximum space
 * available for wrapping will be placed on a line of their own.
 *
 * Released to the public domain in 2001 by Alexander G. M. Smith.  Feel free
 * to improve it and upload a new version to http://www.bebits.com/
 *
 * $Log: AGMSWordWrapper.cpp,v $
 * Revision 1.8  2002/10/28 23:59:25  agmsmith
 * Exclude the less than sign from indent text, since it happens to
 * be the start of an HTML tag and it isn't nice to have a less than
 * sign on EVERY line of text.
 *
 * Revision 1.7  2002/10/28 20:13:19  agmsmith
 * Added tab stops and a list of known abbreviations, and fixed up a few
 * things.
 *
 * Revision 1.6  2001/12/21 22:08:07  agmsmith
 * Single digit numbers at the end of a sentence need two spaces after
 * the period, not one.  But keep a single space for a single letter
 * at the end of a sentence.
 *
 * Revision 1.5  2001/07/15 14:37:39  agmsmith
 * Added compile date string and made it stick in the executable.
 *
 * Revision 1.4  2001/07/15 14:29:42  agmsmith
 * Added version string to the executable.
 *
 * Revision 1.3  2001/07/15 14:17:11  agmsmith
 * Added an understanding of C style comments.
 *
 * Revision 1.2  2001/07/15 02:14:09  agmsmith
 * Largely working now, but needs some special case code for
 * C style comments (don't include the initial slash-star in
 * the indent text).
 *
 * Revision 1.1  2001/07/14 20:27:20  agmsmith
 * Initial revision
 */

#include "MTextAddOn.h" /* For add-on interface. */
#include <String.h>     /* For BString. */
#include <ctype.h>      /* For isalpha etc. */

#define LINE_LENGTH 79
/* Number of characters per line (not bytes since multibyte characters are
allowed in UTF-8).  Words that would make the line longer than this get wrapped
onto the next line.  Change this and recompile if you want a different one. */

#define TAB_SIZE 4
/* There are tab stops every this many characters.  This is currently set to
four for use with the OpenTracker coding guidelines, which are used in many
BeOS projects, and assume tabs every 4 characters.  Of course, my coding style
is indents by 2 spaces, no tabs allowed, to avoid tab size ambiguity problems
like this.  See http://opentracker.sourceforge.net/guidelines.html for more.  I
was unable to find a way of getting the current tab size setting from the
editor environment, thus this constant. */

extern "C" _EXPORT status_t perform_edit (MTextAddOn *EditorCallbackPntr);

static const char VersionString [] = "$Header: /boot/home/agmsmith/Programming/WordWrapper/RCS/AGMSWordWrapper.cpp,v 1.8 2002/10/28 23:59:25 agmsmith Exp $\n";
static const char CompileDateString [] = "Compiled on " __DATE__ " at " __TIME__ ".\n";

/* A list of abbreviations that are never at the end of a sentence.  When these
words are found (case insensitive), the period at the end won't be considered
as ending a sentence, so they will be followed by one space rather than two.
The end of the list is marked by a NULL pointer.  For convenience, it is in
alphabetical order. */

static const char *AbbreviationList [] =
{
  "Dr.",
  "Mr.",
  "Mrs.",
  "Ms.",
  "St.",
  NULL
};



/******************************************************************************
 * This function finds the number of characters long the string is, not the
 * number of bytes.  Here is how Unicode characters are mapped to UTF-8:
 * U-00000000 - U-0000007F: 0xxxxxxx
 * U-00000080 - U-000007FF: 110xxxxx 10xxxxxx
 * U-00000800 - U-0000FFFF: 1110xxxx 10xxxxxx 10xxxxxx
 * U-00010000 - U-001FFFFF: 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
 * U-00200000 - U-03FFFFFF: 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
 * U-04000000 - U-7FFFFFFF: 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
 *
 * The input string is in StringPntr.  The end of the string is pointed out by
 * StringEndPntr, just in case you want to use NUL bytes as characters.  If
 * UseTabStops is false then tabs count as one character, otherwise they count
 * as enough characters to reach the next tab stop, with tab stops placed every
 * TAB_SIZE characters.
 */

int UTFStringLength (
  const char *StringPntr,
  const char *StringEndPntr,
  bool UseTabStops = false)
{
  int Count = 0;
  int Letter;

  while (StringPntr < StringEndPntr)
  {
    Letter = (unsigned char) *StringPntr++;

    if (B_TAB == Letter && UseTabStops)
    {
      Count = (Count + TAB_SIZE) / TAB_SIZE * TAB_SIZE;
    }
    else if (Letter <= 0x7F)
    {
      Count++; /* Just a single ASCII equivalent letter. */
    }
    else if (Letter <= 0xBF)
    {
      /* 0x80 to 0xBF is a middle byte of a multi-byte sequence, shouldn't be
      here at the start of a sequence, just skip it. */
    }
    else if (Letter <= 0xFD)
    {
      /* Start of a multi-byte character. */
      Count++;

      /* Get the number of following bytes.  Well, actually we don't need to,
      but I'll just leave this code here as a comment.
      if (Letter < 0xE0)
        RemainingBytes = 1;
      else if (Letter < 0xF0)
        RemainingBytes = 2;
      else if (Letter < 0xF8)
        RemainingBytes = 3;
      else if (Letter < 0xFC)
        RemainingBytes = 4;
      else
        RemainingBytes = 5;
      */

      /* Skip over the remaining bytes.  They should all be middle bytes, which
      are between 0x80 and 0xBF.  If not, assume that some bytes were missing
      and stop skipping.  Well, the main loop actually does this for us. */
    }
    else
    {
      /* An invalid character, 0xFE or 0xFF, skip it. */
    }
  }

  return Count;
}



/******************************************************************************
 * This is called by the text editor when the user selects the menu item that
 * activates our plug-in.
 */

status_t perform_edit (MTextAddOn *EditorCallbackPntr)
{
  int           AbbreviationIndex;
  const char   *AbbreviationPntr;
  bool          EndOfCStyleComment;
  bool          FirstTimeInput = true;
  bool          FirstTimeOutput = true;
  int           IndentByteLength;
  int           IndentCharacterLength;
  const char   *IndentComparisonPntr;
  const char   *IndentTextEndPntr;
  const char   *IndentTextStartPntr;
  int           Letter;
  int           NewByteLength;
  BString       NewText;
  bool          PreviousWordEndsSentence = false;
  status_t      ReturnCode = -1;
  int32         SelectionEnd;
  const char   *SelectionEndPntr;
  int32         SelectionStart;
  const char   *SelectionStartPntr;
  const char   *StringPntr;
  const char   *TextEndPntr;
  const char   *TextStartPntr;
  int           WordByteLength;
  int           WordCharacterLength;
  const char   *WordEndPntr;
  const char   *WordStartPntr;
  const char   *WordTestPntr;
  int           WrapRemainingSpace;
  int           WrapTotalSpace;

  if (!EditorCallbackPntr->IsEditable ())
    goto ErrorExit; /* Do nothing for read only files. */

  /* A couple of do-nothing calls so that the compiler doesn't eliminate
  the version and compile date strings from the executable. */

  UTFStringLength (VersionString, VersionString);
  UTFStringLength (CompileDateString, CompileDateString);

  /* Find the start and end of the text to be wrapped, extending the user's
  selection to the start of the first line and the end of the last line. */

  EditorCallbackPntr->GetSelection (&SelectionStart, &SelectionEnd);
  TextStartPntr = EditorCallbackPntr->Text ();
  TextEndPntr = TextStartPntr + EditorCallbackPntr->TextLength ();

  StringPntr = TextStartPntr + SelectionStart - 1;
  while (StringPntr >= TextStartPntr && *StringPntr != '\n')
  {
    StringPntr--;
    SelectionStart--;
  }

  StringPntr = TextStartPntr + SelectionEnd;
  while (StringPntr < TextEndPntr && *StringPntr != '\n')
  {
    StringPntr++;
    SelectionEnd++;
  }

  EditorCallbackPntr->Select (SelectionStart, SelectionEnd);

  TextStartPntr = EditorCallbackPntr->Text ();
  TextEndPntr = TextStartPntr + EditorCallbackPntr->TextLength ();
  SelectionStartPntr = TextStartPntr + SelectionStart;
  SelectionEndPntr = TextStartPntr + SelectionEnd;

  /* Find the indent text at the start of the line.  First look for white space
  followed by a slash and a star.  Stop searching at the end of the line. */

  StringPntr = SelectionStartPntr;
  while (StringPntr < SelectionEndPntr)
  {
    Letter = *StringPntr;
    if (isspace (Letter) && (Letter != '\n'))
      StringPntr++;
    else
      break;
  }

  if ((SelectionEndPntr - StringPntr) < 2 /* if no space for a slash-star */ ||
  StringPntr[0] != '/' || StringPntr[1] != '*')
  {
    /* Not the start of a C style comment, so the indent text is all the space
    and punctuation characters up to the actual text or end of line.  Well,
    exclude "<" from punctuation since that's used as the start of HTML tags
    and is annoying if it is considered part of the indent text. */

    StringPntr = SelectionStartPntr;
    while (StringPntr < SelectionEndPntr)
    {
      Letter = *StringPntr;
      if ((isspace (Letter) || (Letter != '<' && ispunct (Letter))) &&
      (Letter != '\n'))
        StringPntr++;
      else
        break;
    }
  }

  IndentByteLength = StringPntr - SelectionStartPntr;
  IndentTextStartPntr = SelectionStartPntr;
  IndentTextEndPntr = IndentTextStartPntr + IndentByteLength;
  IndentCharacterLength =
    UTFStringLength (IndentTextStartPntr, IndentTextEndPntr, true /* tabs */);
  WrapTotalSpace = LINE_LENGTH - IndentCharacterLength;
  if (WrapTotalSpace <= 0)
    goto ErrorExit; /* No space for wrapping text, do nothing. */

  /* Do the wrapping.  Each line's source excludes any text in common with the
  indent text, and gets broken down into words, which then get appended to
  NewText. */

  StringPntr = SelectionStartPntr;
  WrapRemainingSpace = 0;
  while (StringPntr < SelectionEndPntr)
  {
    /* Skip over white space.  Also skip over the start of a line which matches
    the indent text. */

    while (StringPntr < SelectionEndPntr)
    {
      Letter = *StringPntr;
      if (Letter == '\n' || FirstTimeInput)
      {
        if (FirstTimeInput)
          FirstTimeInput = false;
        else
          StringPntr++; /* Skip over the newline character. */

        /* At the start of a line, skip the rest of the text so long as it
        matches the indent text. */

        IndentComparisonPntr = IndentTextStartPntr;
        while (StringPntr < SelectionEndPntr &&
        IndentComparisonPntr < IndentTextEndPntr &&
        *StringPntr == *IndentComparisonPntr)
        {
          StringPntr++;
          IndentComparisonPntr++;
        }
      }
      else if (isspace (Letter))
      {
        /* Skip the spaces between words. */
        StringPntr++;
      }
      else /* Got a non-space letter. */
        break; /* Found start of a word, exit searching loop. */
    }

    /* Find the end of the word - at the next white space. */

    WordStartPntr = StringPntr;
    while (StringPntr < SelectionEndPntr && !isspace (*StringPntr))
      StringPntr++;
    WordEndPntr = StringPntr;
    WordByteLength = WordEndPntr - WordStartPntr;
    if (WordByteLength <= 0)
      continue; /* No word, try again, also will exit if no more data. */
    WordCharacterLength = UTFStringLength (WordStartPntr, WordEndPntr);

    /* Append the word to the new text.  Preceed with a space (or two) if it is
    continuing a previous line, preceed with the indent text if it is starting
    a new line. */

    EndOfCStyleComment = (WordByteLength == 2 &&
      WordStartPntr[0] == '*' && WordStartPntr[1] == '/');

    if (WordCharacterLength +
    (EndOfCStyleComment ? 1 : (PreviousWordEndsSentence ? 2 : 1)) >
    WrapRemainingSpace)
    {
      /* Not enough space, start a new line. */
      if (FirstTimeOutput)
        FirstTimeOutput = false;
      else
        NewText.Append ("\n");
      NewText.Append (IndentTextStartPntr, IndentByteLength);
      NewText.Append (WordStartPntr, WordByteLength);
      WrapRemainingSpace = WrapTotalSpace - WordCharacterLength;
    }
    else
    {
      /* Word fits on line, append to existing words with a space or two to
      separate them.  The star-slash end of a C style comment will always have
      one space between it and the previous word, even if the previous word
      ended in a period. */

      if (EndOfCStyleComment || !PreviousWordEndsSentence)
      {
        NewText.Append (" ");
        WrapRemainingSpace -= 1;
      }
      else /* Starting a new sentence. */
      {
        NewText.Append ("  ");
        WrapRemainingSpace -= 2;
      }
      NewText.Append (WordStartPntr, WordByteLength);
      WrapRemainingSpace -= WordCharacterLength;
    }

    /* Set a flag to show if this word ends a sentence, so that the next word
    can have two spaces between it and the end of sentence.  Consider single
    letter words (not digits or other characters) followed by a period to be an
    abbreviation, so that "Number 1.  Second number 2.  Punctuation +.  Etc."
    space out as sentences but "Alexander G. M. Smith" doesn't get spaced out
    too much :-).  Also look for known abbreviations which end in a period but
    are never at the end of the sentence (like "Mr."). */

    if (WordEndPntr[-1] == '!')
      PreviousWordEndsSentence = true;
    else if (WordEndPntr[-1] == '?')
      PreviousWordEndsSentence = true;
    else if (WordEndPntr[-1] == '.')
    {
      PreviousWordEndsSentence =
        (WordCharacterLength > 2 ||
        (WordCharacterLength == 2 && !isalpha (WordEndPntr[-2])));

      if (PreviousWordEndsSentence)
      {
        /* Check the list of known abbreviations that aren't at the end of a
        sentence, even though the word ends with a period. */

        for (AbbreviationIndex = 0;
        AbbreviationList [AbbreviationIndex] != NULL; AbbreviationIndex++)
        {
          AbbreviationPntr = AbbreviationList [AbbreviationIndex];
          WordTestPntr = WordStartPntr;
          while (*AbbreviationPntr != 0 && WordTestPntr < WordEndPntr)
          {
            if (tolower (*AbbreviationPntr) != tolower (*WordTestPntr))
              break; /* Mismatch, this isn't the word we're looking for. */
            AbbreviationPntr++;
            WordTestPntr++;
          }
          if (WordTestPntr == WordEndPntr && *AbbreviationPntr == 0)
          {
            /* Found the word in our abbreviation list. */
            PreviousWordEndsSentence = false;
            break;
          }
        }
      }
    }
    else
      PreviousWordEndsSentence = false;
  }

  /* Replace the old text with the new text. */

  EditorCallbackPntr->Delete ();
  NewByteLength = NewText.Length();
  if (NewByteLength > 0)
    EditorCallbackPntr->Insert (NewText.String(), NewByteLength);

  ReturnCode = 0; /* Successful if we got this far. */

  /* Clean up. */

ErrorExit:
  return ReturnCode;
}
